/*

    m_main.h

 */

#ifndef __m_main_h
#define __m_main_h


#include "g_main.h"


typedef struct {

    char *data;
    int  len;
    } cFile;

void    m_init(void);         // muss zu allererst aufgerufen werden
void    m_shutdown(void);     // muss zu allerletzt aufgerufen werden


// Speicher

char*    m_alloc(int);         //Reserviert speicher
void     m_free(char*);        // ..
cbitmap *m_allocbitmap (int x, int y, int bpp);


// Fehlermeldungen


void    m_fatalerror(char *); // Fataler Fehler (Abbruch)
void    m_print(char *);      // gibt text in aus..


cbitmap *m_loadbitmap (char *);
void    m_freebitmap (cbitmap *);

cFile*   m_loadfile(char*);
void    m_freefile(cFile*);

#endif
